#!/usr/bin/python
# -*- coding: utf-8 -*-

""" The Test Runner """

import logging
import sys
from pathlib import Path
from typing import List, Dict

# pylint: disable=C0303, R0903, R1721, W0703, W1203

sys.path.insert(0, str(Path("..").absolute()))

# Imports may differ on how and from where PackageTester is invoked.
try:
    from .package_tester_utils import (
        get_project_root,
        PytestResultsParser,
        execute_shell_command,
        ENTRYPOINT,
        TestedPackageInfo,
    )
except ImportError:
    print(">>> FYI: using alternative imports")
    from package_tester_utils import (
        get_project_root,
        PytestResultsParser,
        execute_shell_command,
        ENTRYPOINT,
        TestedPackageInfo,
    )

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class PackageTester:
    """Runs pytest for every package specified.
    This version has been refactored to be more self-contained and
    manageable than previous version. It is invoked from an
    entrypoint rather than a shell script.
    """

    def __init__(self, packages: str, verbose=False):
        self.package_names = packages
        self.verbose = verbose or False
        # to be List[TestPackageInfo]
        self.pytested_packages = []
        # to be List[str]
        self.output = []
        # to be Dict[str, Path]
        self.locations_by_package = {}

    def __str__(self):
        """
        If there's output, don't include me

        :return:
        """
        return ""

    def execute(self):
        """
        This is where the magic happens. Find the packages the user submitted;
        `pytest` each; organize, report, and store results.

        :return:
        """
        self.locations_by_package = PackageFinder(self.package_names).find()

        if len(self.locations_by_package) == 0:
            raise PackageTesterException("One or more packages not found.")

        if len(self.locations_by_package) < len(self.package_names):
            message = "\nOne or more of your packages is ambiguous; "
            message += "please try again with a more qualified package."
            raise PackageTesterException(message)

        self.output.append(
            f"{len(self.locations_by_package)}" " Package locations found:"
        )
        for package, location in self.locations_by_package.items():
            self.output.append(f"\t{package}: {location}")

        self.pytested_packages = self.process_packages()
        self.summarize()
        self.report()

    def process_packages(self) -> List[TestedPackageInfo]:
        """
        Run pytest on each of the packages.

        :return:
        """

        self.make_dev()

        packages_info = []
        for package, location in self.locations_by_package.items():
            package_info = TestedPackageInfo(package, location)
            # cd to package location, -then- run pytest
            command = f'cd {location}; echo "Current directory: $(pwd)"; ' f"pytest -v"
            output = execute_shell_command(command)
            for line in output:
                package_info.test_output.append(line)
            packages_info.append(package_info)

        return packages_info

    @staticmethod
    def make_dev():
        """
        Execute 'make dev' from top level; should have the effect of
        running 'python setup.py develop' in every package that has a
        setup.py

        :return:
        """
        root = get_project_root()
        command = f"cd {str(root)}; make dev"

        output = execute_shell_command(command)

        for line in output:
            assert not line.startswith("Not")
            assert "error" not in line.lower()

    def summarize(self):
        """
        Summarize test results for each package tested.
        :return:
        """

        for package_info in self.pytested_packages:
            result_line = package_info.test_output[-1]
            result = PytestResultsParser(result_line).parse()
            package_info.summary = result

    def report(self):
        """
        Report results.

        :return:
        """

        for package_info in self.pytested_packages:
            if self.verbose:
                details = [f"{line}" for line in package_info.test_output]
                for line in details:
                    print(line)
            print(f"{package_info.summary}")


def build_usage_message() -> str:
    """
    Tell 'em how to use this thing.

    :return:
    """
    message = f"Usage: {ENTRYPOINT} pkg1 [pkg2 pkg3 ... pkg N]\n\n"
    message += "pkg arguments are names of workspaces packages to be tested;\n"
    message += 'e.g., "src.apps.cli.executables.datafetcher" or\n'
    message += '"shared.channels"'
    return message


class PackageFinder:
    """ Finds paths corresponding to each submitted package name. """

    def __init__(self, packages: List[str]):

        # make sure we have something to work with
        failure_message = "One or more package names required\n" + build_usage_message()
        if (not packages) or len(packages) == 0:
            raise PackageTesterException(failure_message)
        for package in packages:
            if package is None:
                raise PackageTesterException(failure_message)

        self.packages = packages

    def find(self) -> Dict[str, Path]:
        """
        Get each package's path.

        :return:
        """
        locations_by_package = {}

        for package in self.packages:
            for file in get_project_root().rglob(package.replace(".", "/")):
                # we don't want the built packages; just the source.
                if "build" not in str(file):
                    locations_by_package[package] = file
                    break

        if len(locations_by_package) != len(self.packages):
            raise PackageTesterException(
                f"expected {len(self.packages)} "
                f"packages; got "
                f"{len(locations_by_package)}"
            )

        return locations_by_package


class PackageTesterException(Exception):
    """ Throw this whenever things go south. """


def main() -> PackageTester:
    """
    `package-tester` entrypoint invokes this to launch the tester.

    :return:
    """

    args = sys.argv[1:]
    if args is None or len(args) == 0:
        failure_message = "One or more package names required\n" + build_usage_message()
        logger.error(failure_message)
        raise PackageTesterException(failure_message)

    # let verbose be first OR last arg
    package_names = []
    verbose = False
    for arg in args:
        if arg in ("-v", "--verbose"):
            verbose = True
        else:
            package_names.append(arg)

    tester = PackageTester(package_names, verbose)
    tester.execute()

    return tester


if __name__ == "__main__":
    main()
